# -*- coding: utf-8 -*-
"""Unit Tests for `bottle_neck.middleware` module.
"""


import pytest
import bottle_neck.middleware as middleware


@pytest.fixture(scope='module')
def mock_middleware():
    """pytest Fixture for 'bottle_neck.middleware.BaseMiddleware' subclass.
    """

    class MockMiddleware(middleware.BaseMiddleware):
        def __call__(self, e, h):
            e['PATH_INFO'] = e['PATH_INFO'].rstrip('/') + '#'
            return self.app(e, h)

    return MockMiddleware


def test_middleware_getattr(mock_middleware, mock_app):
    """Testing 'bottle_neck.middleware.BaseMiddleware' app wrapping method.
    """

    mock_app = mock_middleware(mock_app)

    assert mock_app.route


def test_middleware_getattr_multiple(mock_middleware, mock_app):
    """Testing 'bottle_neck.middleware.BaseMiddleware' app wrapping method
    for multiple wraps.
    """
    mock_app = mock_middleware(mock_middleware(mock_middleware(mock_app)))

    assert mock_app.route



# -*- coding: utf-8 -*-
"""Unit Tests for `bottle_neck.routing` module.
"""


import pytest
import bottle_neck.routing as routing
import bottle_neck.cbv as handler


@pytest.fixture(scope='module')
def mock_router():
    """pytest fixture for `bottle_neck.routing.Router` class
    """

    return routing.Router()


@pytest.fixture(scope="module")
def mock_handler():
    """Pytest fixture for `bottle_neck.handlers.BaseHandler` subclass.
    """

    class MockHandler(handler.BaseHandler):
        """Web handler fixture class.
        """
        cors_enabled = True
        base_endpoint = '/mock'

        def get(self, mock_id):
            return mock_id

    return MockHandler


def test_router_register_handler_fn_pass(mock_router):
    """Test `bottle_neck.routing.Router.register_handler` for function-based
    handler pass.
    """

    def fn():
        pass

    mock_router.register_handler(fn, entrypoint='/', methods=('GET', ))

    assert len(mock_router) == 1


def test_router_register_handler_fail(mock_router):
    """Test `bottle_neck.routing.Router.register_handler` error handling.
    """

    with pytest.raises(routing.RouteError):
        mock_router.register_handler('handler', entrypoint='/')


def test_router_register_handler_cbv_pass(mock_router, mock_handler):
    """Test `bottle_neck.routing.Router.register_handler` for class-based
    handler pass.
    """

    mock_router.register_handler(mock_handler, entrypoint='/api')

    assert len(mock_router) == 2


def test_router_mount_pass(mock_router, mock_app):
    """Test `bottle_neck.routing.Router.mount` method.
    """
    init_mounts = len(mock_app.routes)
    mock_router.mount(mock_app)

    assert len(mock_app.routes) > init_mounts



# -*- coding: utf-8 -*-
"""Bottle.py Application Plugin utilities.

Provides some base functionality for creating application-scope plugins for
`bottle.py`.
"""

from __future__ import absolute_import

__author__ = 'pav'
__date__ = '2015-12-9'
__all__ = ['BasePlugin', 'WrapErrorPlugin']


import abc
import bottle
from bottle_neck import __version__
import six

version = tuple(map(int, __version__.split('.')))


@six.add_metaclass(abc.ABCMeta)
class BasePlugin(object):
    """'bottle.py' Plugin API version 2 interface.

    Provides a concrete Plugin Base class for creating bottle.py plugins
    """

    name = None

    api = 2

    def setup(self, app):  # pragma: no cover
        """Make sure that other installed plugins don't affect the same
        keyword argument and check if metadata is available.
        """
        for other in app.plugins:
            if not isinstance(other, self.__class__):
                continue
            if other.keyword == self.keyword:
                raise bottle.PluginError(
                    "Found other plugin registered as {} (non-unique keyword)."
                )

    @abc.abstractmethod
    def apply(self, callback, context):  # pragma: no cover
        pass


class WrapErrorPlugin(BasePlugin):
    """Middleware class that catches `bottle.HTTPError` exceptions and returns
    default HTTP status code 200 using `bottle_neck.response.WSResponse` class
    for error wrapping.
    """

    def __init__(self, keyword, error_wrapper_cls):  # pragma: no cover
        self.error_wrapper = error_wrapper_cls
        self.keyword = keyword

    def apply(self, callback, context):  # pragma: no cover
        """Apply the HTTPError wrapper to the callback.
        """

        def wrapper(*args, **kwargs):
            try:
                return callback(*args, **kwargs)
            except bottle.HTTPError as error:
                return self.error_wrapper.from_status(
                    status_line=error.status_line,
                    msg=error.body
                )

        return wrapper



# -*- coding: utf-8 -*-
#
#    Copyright (C) 2015  Papavassiliou Vassilis
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""`bottle-neck` project.

Provides useful utilities for creating web-services with bottle.py
micro-framework.
"""

from __future__ import absolute_import

__author__ = 'Papavassiliou Vassilis'
__date__ = '2016-1-26'
__version__ = '1.13'

__all__ = ['BaseHandler', 'BaseHandlerPlugin', 'route_method', 'plugin_method',
           'HandlerError', 'HandlerPluginError', 'HandlerHTTPMethodError',
           'WSResponse', 'WSResponseError', 'Router', 'Route',
           'RouteError', 'BasePlugin', 'WrapErrorPlugin', 'BaseMiddleware',
           'StripPathMiddleware', 'cors_enable_hook', 'strip_path_hook',
           'paginator', 'version']


from bottle_neck.cbv import (
    BaseHandler, BaseHandlerPlugin, route_method, plugin_method, HandlerError,
    HandlerPluginError, HandlerHTTPMethodError
)
from bottle_neck.response import (WSResponse, WSResponseError)
from bottle_neck.routing import (RouteError, Router, Route)
from bottle_neck.plugins import (BasePlugin, WrapErrorPlugin)
from bottle_neck.middleware import (BaseMiddleware, StripPathMiddleware)
from bottle_neck.webapi import (cors_enable_hook, strip_path_hook, paginator)


version = tuple(map(int, __version__.split('.')))



# -*- coding: utf-8 -*-
"""Unified Web Service Response Interface for JSON views.

Provides some base classes that implement a strict skeleton for
API response using all HTTP method available. The concept is to
have a unified response object for any web-service that is
framework-agnostic.


.. note::
    In beta version only JSON renderer is supported. Must implement
    additional renderer classes (SOAP, XML, YAML).
"""


from bottle_neck import __version__
import bottle
import collections
import six

__author__ = 'Papavassiliou Vassilis'
__all__ = ['WSResponse', 'WSResponseError']


version = tuple(map(int, __version__.split('.')))

HTTP_CODES = (
    (200, 'OK'),
    (201, 'Created'),
    (304, 'Not Modified'),
    (400, 'Bad Request'),
    (401, 'Unauthorized'),
    (403, 'Forbidden'),
    (404, 'Not Found'),
    (405, 'Method Not Allowed'),
    (501, 'Not Implemented'),
    (503, 'Service Unavailable')
)


class WSResponseError(Exception):
    """Base module exception
    """
    pass


class WSResponse(object):
    """Base web service response class.

    WSResponse class provides a unified API for HTTP responses.

    The Response body skeleton looks like this::

         {
            "status_code": 200,
            "status_txt": "OK",
            "response_data": {
                "id": 65234
                "username": "pav"
                "email": "pav@geotagaeroview.com"
            },
            "errors": []

        }

    Attributes:
        status_code (int): Response HTTP status code.
        data (object): Response key/value data.
        errors (dict): Response key/value errors.

    Examples:
        >>> response = WSResponse(
        ...                 status_code=200,
        ...                 data='Hi'
        ...             )
        ...
        >>> response.to_json
        OrderedDict([('status_code', 200), ('status_text', 'OK'), ('data', 'Hi'), ('errors', [])])

    """
    __slots__ = ['status_code', 'data', 'errors']

    expose_status = False
    response = bottle.response

    def __init__(self, status_code=200, data=None, errors=None):
        if status_code not in dict(HTTP_CODES) or\
                not isinstance(errors, six.string_types + (list, tuple, type(None),)):
            raise WSResponseError('Invalid Response initialization.')
        self.status_code = status_code
        self.data = data

        if isinstance(errors, (six.string_types, )):
            errors = [errors]

        self.errors = errors

    def __repr__(self):  # pragma: no cover
        return "WebService Response: status={}, data={}".format(
            self.status_code, str(self.data)
        )

    def __eq__(self, other):  # pragma: no cover
        assert isinstance(other, WSResponse), 'Invalid Type for eq operator.'
        return self.status_code == other.status_code and \
            self.data == self.data

    __str__ = __repr__

    @classmethod
    def from_status(cls, status_line, msg=None):
        """Returns a class method from bottle.HTTPError.status_line attribute.
        Useful for patching `bottle.HTTPError` for web services.

        Args:
            status_line (str):  bottle.HTTPError.status_line text.
            msg: The message data for response.

        Returns:
            Class method based on status_line arg.

        Examples:
            >>> status_line = '401 Unauthorized'
            >>> error_msg = 'Get out!'
            >>> resp = WSResponse.from_status(status_line, error_msg)
            >>> resp['errors']
            ['Get out!']
            >>> resp['status_text']
            'Unauthorized'
        """
        method = getattr(cls, status_line.lower()[4:].replace(' ', '_'))
        return method(msg)

    @classmethod
    def ok(cls, data):
        """Shortcut API for HTTP 200 `OK` response.

        Args:
            data (dict): Response key/value data.

        Returns
            WSResponse Instance.
        """
        return cls(
            status_code=200,
            data=data
        ).to_json

    @classmethod
    def created(cls, data=None):
        """Shortcut API for HTTP 201 `Created` response.

        Args:
            data (dict): Response key/value data.

        Returns:
            WSResponse Instance.
        """
        if cls.expose_status:
            cls.response.content_type = 'application/json'
            cls.response._status_line = '201 Created'

        return cls(201, data=data).to_json

    @classmethod
    def not_modified(cls, errors=None):
        """Shortcut API for HTTP 304 `Not Modified` response.

        Args:
            errors (list): Response key/value data.

        Returns:
            WSResponse Instance.
        """
        if cls.expose_status:
            cls.response.content_type = 'application/json'
            cls.response._status_line = '304 Not Modified'

        return cls(304, None, errors).to_json

    @classmethod
    def bad_request(cls, errors=None):
        """Shortcut API for HTTP 400 `Bad Request` response.

        Args:
            errors (list): Response key/value data.

        Returns:
            WSResponse Instance.
        """
        if cls.expose_status:
            cls.response.content_type = 'application/json'
            cls.response._status_line = '400 Bad Request'

        return cls(400, errors=errors).to_json

    @classmethod
    def unauthorized(cls, errors=None):
        """Shortcut API for HTTP 401 `Unauthorized` response.

        Args:
            errors (list): Response key/value data.

        Returns:
            WSResponse Instance.
        """
        if cls.expose_status:
            cls.response.content_type = 'application/json'
            cls.response._status_line = '401 Unauthorized'

        return cls(401, errors=errors).to_json

    @classmethod
    def forbidden(cls, errors=None):
        """Shortcut API for HTTP 403 `Forbidden` response.

        Args:
            errors (list): Response key/value data.

        Returns:
            WSResponse Instance.
        """
        if cls.expose_status:
            cls.response.content_type = 'application/json'
            cls.response._status_line = '403 Forbidden'

        return cls(403, errors=errors).to_json

    @classmethod
    def not_found(cls, errors=None):
        """Shortcut API for HTTP 404 `Not found` response.

        Args:
            errors (list): Response key/value data.

        Returns:
            WSResponse Instance.
        """
        if cls.expose_status:
            cls.response.content_type = 'application/json'
            cls.response._status_line = '404 Not Found'

        return cls(404, None, errors).to_json

    @classmethod
    def method_not_allowed(cls, errors=None):
        """Shortcut API for HTTP 405 `Method not allowed` response.

        Args:
            errors (list): Response key/value data.

        Returns:
            WSResponse Instance.
        """
        if cls.expose_status:
            cls.response.content_type = 'application/json'
            cls.response._status_line = '405 Method Not Allowed'

        return cls(405, None, errors).to_json

    @classmethod
    def not_implemented(cls, errors=None):
        """Shortcut API for HTTP 501 `Not Implemented` response.

        Args:
            errors (list): Response key/value data.

        Returns:
            WSResponse Instance.
        """
        if cls.expose_status:
            cls.response.content_type = 'application/json'
            cls.response._status_line = '501 Not Implemented'

        return cls(501, None, errors).to_json

    @classmethod
    def service_unavailable(cls, errors=None):
        """Shortcut API for HTTP 503 `Service Unavailable` response.

        Args:
            errors (list): Response key/value data.

        Returns:
            WSResponse Instance.
        """
        if cls.expose_status:
            cls.response.content_type = 'application/json'
            cls.response._status_line = '503 Service Unavailable'

        return cls(503, None, errors).to_json

    @property
    def to_json(self):
        """Short cut for JSON response service data.

        Returns:
            Dict that implements JSON interface.
        """

        web_resp = collections.OrderedDict()

        web_resp['status_code'] = self.status_code
        web_resp['status_text'] = dict(HTTP_CODES).get(self.status_code)
        web_resp['data'] = self.data
        web_resp['errors'] = self.errors or []

        return web_resp


if __name__ == '__main__':  # pragma: no cover
    import doctest
    doctest.testmod()



# -*- coding: utf-8 -*-
"""CBV for bottle.py application instances.

Provides a base class for creating class-based web handlers for bottle.py
application instances with application routing mechanism.
"""

from __future__ import absolute_import

__author__ = "Papavassiliou Vassilis"
__date__ = "2015-11-29"
__all__ = ['BaseHandler', 'HandlerMeta', 'route_method', 'plugin_method',
           'HandlerError', 'HandlerHTTPMethodError', 'HandlerPluginError',
           'BaseHandlerPlugin']


from bottle_neck import __version__
import bottle
import functools
import inspect
import six
import re


version = tuple(map(int, __version__.split('.')))

DEFAULT_ROUTES = ("get", "put", "post", "delete", "patch", "options")

CORS_ROUTES = ("put", "post", "delete", "patch")

PLUGIN_SCOPE = (
    (False, 'plugins'),
    (True, 'global_plugins')
)


class HandlerError(Exception):
    """Base module Exception class.
    """
    pass


class HandlerHTTPMethodError(HandlerError):
    """Raises for invalid HTTP method declaration.
    """
    pass


class HandlerPluginError(HandlerError):
    """Raises when a handler plugin error occurs.
    """
    pass


class ClassPropertyDescriptor(object):
    """ClassProperty Descriptor class.

    Straight up stolen from stack overflow Implements class level property
    non-data descriptor.
    """

    def __init__(self, fget, fset=None):
        self.fget = fget
        self.fset = fset

    def __get__(self, obj, cls=None):
        if cls is None:  # pragma: no cover
            cls = type(obj)
        return self.fget.__get__(obj, cls)()


def classproperty(func):
    """classproperty decorator.
    Using this decorator a class can have a property. Necessary for properties
    that don't need instance initialization. Works exactly the same as a
    normal property.

    Examples:
        >>> class MyClass(object):
        ...     @classproperty
        ...     def my_prop(self):
        ...         return self.__name__ + ' class'
        ...
        >>> MyClass.my_prop
        'MyClass class'
    """
    if not isinstance(func, (classmethod, staticmethod)):
        func = classmethod(func)
    return ClassPropertyDescriptor(func)


def cached_classproperty(fun):
    """A memorization decorator for class  properties.

    It implements the above `classproperty` decorator, with
    the difference that the function result is computed and attached
    to class as direct attribute. (Lazy loading and caching.)
    """
    @functools.wraps(fun)
    def get(cls):
        try:
            return cls.__cache[fun]
        except AttributeError:
            cls.__cache = {}
        except KeyError:  # pragma: no cover
            pass
        ret = cls.__cache[fun] = fun(cls)
        return ret
    return classproperty(get)


def plugin_method(*plugin_names):
    """Plugin Method decorator.
    Signs a web handler function with the plugins to be applied as attributes.

    Args:
        plugin_names (list): A list of plugin callable names

    Returns:
        A wrapped handler callable.

    Examples:
        >>> @plugin_method('json', 'bill')
        ... def method():
        ...     return "Hello!"
        ...
        >>> print method.json
        True
        >>> print method.bill
        True

    """
    def wrapper(callable_obj):
        for plugin_name in plugin_names:
            if not hasattr(callable_obj, plugin_name):
                setattr(callable_obj, plugin_name, True)
        return callable_obj
    return wrapper


def route_method(method_name, extra_part=False):
    """Custom handler routing decorator.
    Signs a web handler callable with the http method as attribute.

    Args:
        method_name (str): HTTP method name (i.e GET, POST)
        extra_part (bool): Indicates if wrapped callable name should be a part
                           of the actual endpoint.

    Returns:
        A wrapped handler callable.

    examples:
        >>> @route_method('GET')
        ... def method():
        ...     return "Hello!"
        ...
        >>> method.http_method
        'GET'
        >>> method.url_extra_part
        None
    """
    def wrapper(callable_obj):
        if method_name.lower() not in DEFAULT_ROUTES:
            raise HandlerHTTPMethodError(
                'Invalid http method in method: {}'.format(method_name)
            )

        callable_obj.http_method = method_name.upper()

        callable_obj.url_extra_part = callable_obj.__name__ if extra_part\
            else None

        return classmethod(callable_obj)
    return wrapper


class BaseHandlerPlugin(object):
    """
    """
    def __init__(self, callable_object, *args, **kwargs):  # pragma: no cover
        self._wrapped = callable_object
        self._args = args
        self._kwargs = kwargs
        self.__doc__ = callable_object.__doc__

    @cached_classproperty
    def func_name(self):
        cls_name = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', self.__name__)
        return re.sub('([a-z0-9])([A-Z])', r'\1_\2', cls_name).lower()

    def apply(self, *args, **kwargs):  # pragma: no cover
        raise NotImplementedError("Must Override `apply` method")

    def __call__(self, *args, **kwargs):  # pragma: no cover
        return self.apply(*args, **kwargs)


class HandlerMeta(type):
    """BaseHandler metaclass

    Provides meta functionality for ensuring that `http_method` don't require
    an instance and test and that derived classes implements `Singleton`.
    """
    _instances = {}

    def __new__(mcs, name, bases, attrs):

        http_methods = [attr for attr in attrs if
                        attr.lower() in DEFAULT_ROUTES]

        for handler in http_methods:
            attrs[handler] = classmethod(attrs[handler])

        return super(HandlerMeta, mcs).__new__(mcs, name, bases, attrs)

    def __call__(cls, *args):
        """Enforce Singleton class initiation
        """
        if cls not in cls._instances:
            cls._instances[cls] = super(
                HandlerMeta, cls
            ).__call__(*args)
        return cls._instances[cls]


@six.add_metaclass(HandlerMeta)
class BaseHandler(object):
    """Base Handler class for implementing Class-Based Web handler for bottle.

    Subclass `BaseHandler` in order to its API for application routing, and
    implement any of the known http method:

    Class attributes:
        base_endpoint (str): The handler endpoint prefix.
        cors_enabled (bool): Indicates if CORS is enabled.
        plugins (dict): A key/value mapping of available plugins.
        global_plugins (dict): A key/value mapping default applying plugins.
    """
    request = bottle.request
    response = bottle.response
    base_endpoint = '/'
    cors_enabled = True
    plugins = dict()
    global_plugins = dict()

    @classmethod
    def add_plugin(cls, plugin_callables, global_scope=False):
        """Register a new plugin in handler.

        Args:
            plugin_callables (list): A list of plugin callables.
            global_scope (bool): Indicates The scope of the plugin.

        Returns:
            Class instance.
        """
        repo = getattr(cls, dict(PLUGIN_SCOPE)[global_scope])

        for plugin_callable in plugin_callables:

            if hasattr(plugin_callable, 'func_name'):
                repo[plugin_callable.func_name] = plugin_callable
            else:  # pragma: no cover
                repo[plugin_callable.__name__] = plugin_callable

        return cls

    @classmethod
    def register_app(cls, application):
        """Register class view in bottle application.

        Args:
            application (instance): A bottle.Bottle() instance.

        Returns:
            Class instance.
        """

        if cls is BaseHandler:
            raise HandlerError("Cant register a `BaseHandler` class instance")

        routes = cls._get_http_members()

        router = getattr(application, "route")

        for func_name, func_callable in routes:
            method_args = inspect.getargspec(func_callable)[0]
            method_args.remove('self')

            if hasattr(func_callable, 'http_method'):
                http_method = func_callable.http_method
                url_extra_part = func_callable.url_extra_part
            else:
                http_method = func_name
                url_extra_part = None

            applied_plugins = [pl for pl in cls.plugins
                               if hasattr(func_callable, pl)]

            for plugin in applied_plugins:  # pragma: no cover
                try:
                    func_callable = cls.plugins[plugin](func_callable)
                except TypeError as error:
                    raise HandlerPluginError(error.args)

            for global_plugin in cls.global_plugins:
                func_callable = cls.global_plugins[global_plugin](
                    func_callable
                )

            router(
                cls._build_routes(method_args, url_extra_part),
                method=[http_method.upper()]
            )(func_callable)

        if cls.cors_enabled:
            cls_desc = cls.__doc__ or ''
            options_data = {
                "handler": {"name": cls.__name__, "desc": cls_desc.strip()},
                "http_methods": [r[0] for r in routes]
            }

            router(
                cls.base_endpoint,
                method=["OPTIONS"]
            )(lambda: options_data)

            router(
                "{}/<url:re:.+>".format(cls.base_endpoint).replace("//", "/"),
                method=["OPTIONS"]
            )(lambda url: options_data)

        return cls

    @classmethod
    def _build_routes(cls, method_args, url_extra_part=None):
        """Create bottle route for a handler http method."""

        prefix = '/{}'.format(url_extra_part) if url_extra_part else ''

        prefix += '/:' if method_args else ''

        endpoint = cls.base_endpoint + prefix + '/:'.join(method_args)

        return endpoint.replace("//", '/')

    @classmethod
    def _get_http_members(cls):
        """Filter all `http` specific method from handler class.
        """
        return [member for member in
                inspect.getmembers(cls, predicate=inspect.ismethod)
                if cls._check_http_member(member)]

    @staticmethod
    def _check_http_member(member):
        """Checks if a class method has HTTP info.
        """
        return member[0].lower() in DEFAULT_ROUTES or \
            hasattr(member[1], 'http_method')


if __name__ == '__main__':  # pragma: no cover
    import doctest
    doctest.testmod()



from distutils.core import setup

setup(
    name='bottle-neck',
    version='1.14',
    packages=['bottle_neck'],
    url='https://github.com/agile4you/bottle-neck.git',
    license='GLPv3',
    author='Papavassiliou Vassilis',
    author_email='vpapavasil@gmail.com',
    description='Web services with bottle.py made easy!',
    extras_require={
        'test': ['pytest', 'bottle', 'six'],
    }
)
